import logging

from .const import ZoneInput, ZoneStatus, ZoneType
from .spc_error import SpcError
from .utils import _load_enum

_LOGGER = logging.getLogger(__name__)


class Zone:
    """Represents a SPC zone."""

    def __init__(self, bridge, area, zone_data):
        self._bridge = bridge
        self._id = zone_data.get("id")
        self._name = zone_data.get("name")
        self._area = area
        self._http_client = bridge._http_client
        self._type = _load_enum(ZoneType, zone_data.get("type", 0))
        self._values = {
            "input": _load_enum(ZoneInput, zone_data.get("input", 0)),
            "status": _load_enum(ZoneStatus, zone_data.get("status", 0)),
        }

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def values(self):
        return self._values

    def change_values(self, values) -> list:
        changed_values = []

        if values.get("input") is not None:
            new_input = _load_enum(ZoneInput, values["input"])
            if new_input != self._values["input"]:
                self._values["input"] = new_input
                changed_values.append("input")

        if values.get("status") is not None:
            new_status = _load_enum(ZoneStatus, values["status"])
            if new_status != self._values["status"]:
                self._values["status"] = new_status
                changed_values.append("status")

        return changed_values

    @property
    def state(self) -> bool:
        return self._values["input"] == ZoneInput.OPEN

    @property
    def input(self):
        return self._values["input"]

    @property
    def alarm_status(self):
        _alarm_status = {
            "intrusion": False,
            "fire": False,
            "tamper": False,
            "problem": False,
        }
        zone_input = self._values["input"]
        zone_status = self._values["status"]
        zone_type = self._type

        if zone_status == ZoneStatus.TAMPER:
            _alarm_status["tamper"] = True
        elif zone_status == ZoneStatus.ALARM:
            if (
                zone_type == ZoneType.ALARM
                or zone_type == ZoneType.ENTRY_EXIT
                or zone_type == ZoneType.GLASSBREAK
                or zone_type == ZoneType.ENTRY_EXIT_2
            ):
                _alarm_status["intrusion"] = True
            elif zone_type == ZoneType.FIRE:
                _alarm_status["fire"] = True
            elif zone_type == ZoneType.TAMPER:
                _alarm_status["tamper"] = True

        if zone_input != ZoneInput.CLOSED and zone_input != ZoneInput.OPEN:
            _alarm_status["problem"] = True

        return _alarm_status

    @property
    def intrusion(self) -> bool:
        return self.alarm_status.get("intrusion", False)

    @property
    def fire(self) -> bool:
        return self.alarm_status.get("fire", False)

    @property
    def tamper(self) -> bool:
        return self.alarm_status.get("tamper", False)

    @property
    def problem(self) -> bool:
        return self.alarm_status.get("problem", False)

    @property
    def inhibited(self) -> bool:
        return self._values["status"] == ZoneStatus.INHIBIT

    @property
    def isolated(self) -> bool:
        return self._values["status"] == ZoneStatus.ISOLATE

    async def async_command(self, command, code) -> dict:
        if code is None:
            return SpcError(54).error

        username = None
        password = None
        username, password = self._bridge.get_user_credentials(code)
        if username is None or password is None:
            return SpcError(54).error

        return await self._http_client.async_command_zone(
            command, self._id, username, password
        )



import logging

from .const import ArmMode
from .spc_error import SpcError
from .utils import _load_enum

_LOGGER = logging.getLogger(__name__)


class Area:
    """Represents a SPC area."""

    def __init__(self, bridge, area_data):
        self._bridge = bridge
        self._id = area_data.get("id")
        self._name = area_data.get("name")
        self._a_enabled = area_data.get("a_enabled")
        self._a_name = area_data.get("a_name")
        self._b_enabled = area_data.get("b_enabled")
        self._b_name = area_data.get("b_name")
        self._exittime = area_data.get("exittime", 0)
        self._entrytime = area_data.get("entrytime", 0)
        self._http_client = bridge._http_client
        self._values = {
            "mode": _load_enum(ArmMode, area_data.get("mode")),
            "set_user": area_data.get("set_user"),
            "unset_user": area_data.get("unset_user"),
            "changed_by": "",
            "pending_exit": False,
        }
        self.zones = None

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def a_enabled(self):
        return self._a_enabled

    @property
    def a_name(self):
        return self._a_name

    @property
    def b_enabled(self):
        return self._b_enabled

    @property
    def b_name(self):
        return self._b_name

    @property
    def exittime(self):
        return self._exittime

    @property
    def entrytime(self):
        return self._entrytime

    @property
    def mode(self):
        return self._values["mode"]

    @mode.setter
    def mode(self, mode):
        self._values["mode"] = _load_enum(ArmMode, mode)

    @property
    def changed_by(self):
        return self._values["changed_by"]

    @property
    def set_user(self):
        return self._values["set_user"]

    @property
    def unset_user(self):
        return self._values["unset_user"]

    @property
    def pending_exit(self):
        return self._values["pending_exit"]

    @property
    def values(self):
        return self._values

    def change_values(self, values) -> list:
        changed_values = []

        if values.get("mode") is not None:
            new_mode = _load_enum(ArmMode, values["mode"])
            if new_mode != self._values["mode"]:
                self._values["mode"] = new_mode
                changed_values.append("mode")

        if values.get("changed_by") is not None:
            self._values["changed_by"] = values["changed_by"]
            changed_values.append("changed_by")

        if values.get("set_user") is not None:
            self._values["set_user"] = values["set_user"]
            changed_values.append("set_user")

        if values.get("unset_user") is not None:
            self._values["unset_user"] = values["unset_user"]
            changed_values.append("unset_user")

        if values.get("pending_exit") is not None:
            self._values["pending_exit"] = values["pending_exit"]
            changed_values.append("pending_exit")

        return changed_values

    @property
    def alarm_status(self):
        _alarm_status = {
            "intrusion": False,
            "fire": False,
            "tamper": False,
            "problem": False,
            "verified": False,
        }

        for zone in self.zones:
            zone_alarm_status = zone.alarm_status
            for k in zone_alarm_status:
                if zone_alarm_status[k] is True:
                    _alarm_status[k] = True
                    if k == "intrusion" and _alarm_status[k] is True:
                        _alarm_status["verified"] = True
        return _alarm_status

    @property
    def intrusion(self) -> bool:
        return self.alarm_status.get("intrusion", False)

    @property
    def fire(self) -> bool:
        return self.alarm_status.get("fire", False)

    @property
    def tamper(self) -> bool:
        return self.alarm_status.get("tamper", False)

    @property
    def problem(self) -> bool:
        return self.alarm_status.get("problem", False)

    @property
    def verified(self) -> bool:
        return self.alarm_status.get("verified", False)

    async def async_command(self, command, code) -> int:
        if code is None:
            return SpcError(54).error

        username = None
        password = None
        username, password = self._bridge.get_user_credentials(code)
        if username is None or password is None:
            return SpcError(54).error

        if command == "clear_alerts":
            return await self._http_client.async_command_clear_alerts(
                username, password
            )
        else:
            err = await self._http_client.async_command_area(
                command, self._id, username, password
            )
            if command == "set_delayed" and err["code"] == 0:
                self._bridge.set_value("area", self._id, {"pending_exit": True})
            else:
                self._bridge.set_value("area", self._id, {"pending_exit": False})
            return err

    async def async_get_arm_status(self, arm_mode) -> dict:
        return await self._http_client.async_get_area_arm_status(arm_mode, self._id)



class SpcError:
    ERROR_MAP = {
        0: "OK",
        1: "OK (limited)",
        10: "ERROR: Probably already isolated or isolated",
        11: "ERROR: Unknown",
        12: "ERROR: Missing ID",
        13: "ERROR: Invalid ID",
        14: "ERROR: Unknown Tag",
        15: "ERROR: Memory Full",
        16: "ERROR: Invalid Data",
        17: "ERROR: Missing Data",
        18: "ERROR: Invalid CRC",
        19: "ERROR: Invalid Length",
        20: "ERROR: Not ready",
        21: "ERROR: Invalid Sequence No",
        22: "ERROR: Invalid Decryption",
        23: "ERROR: Invalid Connection Details",
        24: "ERROR: Invalid Username",
        25: "ERROR: Invalid Password",
        40: "ERROR: Generic check failed",
        50: "ERROR: Active",
        51: "ERROR: Inactive",
        52: "ERROR: Invalid User",
        53: "ERROR: Command only possible in Normal door mode",
        54: "ERROR: Authentication Failed",
        55: "ERROR: Engineer Not Authorized",
        56: "ERROR: Invalid Name",
        57: "ERROR: Invalid Profile",
        58: "ERROR: Invalid Site Code",
        59: "ERROR: Invalid PIN",
        60: "ERROR: Duplicate",
        61: "ERROR: Invalid Card Number",
        62: "ERROR: In use",
        63: "ERROR: Global ID in use",
        64: "ERROR: Global Data Protected",
        65: "ERROR: No Rights",
        66: "ERROR: System Set",
        67: "ERROR: Cannot delete",
        68: "ERROR: Cannot delete last",
        69: "ERROR: Date",
        70: "ERROR: Calendar",
        71: "ERROR: Area",
        72: "ERROR: Door",
        73: "ERROR: Web password not enabled",
        74: "ERROR: Null data",
        75: "ERROR: Bad Command",
        76: "ERROR: Pin Expired",
        77: "ERROR: Command is blocked, probably due to open zones",
        78: "ERROR: Not allowed in Engineer mode",
        79: "ERROR: Cannot delete default profile",
        80: "ERROR: Cannot edit default profile",
        100: "ERROR: XML - Buffer Fail",
        101: "ERROR: XML - Bad Format",
        102: "ERROR: XML - Bad Data",
        103: "ERROR: XML - Unknown Tag",
        104: "ERROR: XML - Compulsory Parameter Not Found",
        120: "ERROR: File - Fail",
        121: "ERROR: File - No Space",
        122: "ERROR: File - Not Found",
        123: "ERROR: File - Header",
        124: "ERROR: File - Flash",
        125: "ERROR: File - Flash Verify",
        126: "ERROR: File - Flash Erase",
        140: "ERROR: HTTP - Compulsory Parameter Not Found",
        160: "ERROR: SAM - WD Output",
        998: "ERROR: Communication error",
        999: "ERROR: Unknown error",
    }

    def __init__(self, code: int) -> None:
        self._code = code
        self._message = self.ERROR_MAP.get(code, "ERROR: Unknown error")

    @property
    def message(self) -> str:
        return self._message

    @property
    def error(self) -> dict:
        return {"code": self._code, "message": self._message}



"""SPC Websocket client"""

import json
import logging

from .websocket import AIOWSClient

_LOGGER = logging.getLogger(__name__)


class SpcWsClient:
    """SPC Websocket Client Class"""

    def __init__(
        self, bridge, gw_ip_address, gw_port, credentials, loop, session, http_client
    ):
        """Initialize the client."""
        self._bridge = bridge
        self._gw_ip_address = gw_ip_address
        self._gw_port = gw_port
        self._credentials = credentials
        self._loop = loop
        self._session = session
        self._set_value = bridge.set_value
        self._http_client = http_client
        self._websocket = None

    def ws_start(self):
        """Connect websocket to SPC Web Gateway."""
        _LOGGER.debug(
            "Start websocket communication with SPC Bridge: %s:%i",
            self._gw_ip_address,
            self._gw_port,
        )
        proto = "ws"
        params = ""
        if len(self._credentials["ws_username"]) and len(
            self._credentials["ws_password"]
        ):
            proto = "wss"
            params = f"?username={self._credentials["ws_username"]}&password={self._credentials["ws_password"]}&flexc_poll_events=1"

        url = f"{proto}://{self._gw_ip_address}:{self._gw_port}/ws/spc{params}"
        self._websocket = AIOWSClient(
            loop=self._loop,
            session=self._session,
            url=url,
            async_callback=self._async_ws_handler,
        )
        self._websocket.start()

    def ws_stop(self):
        """Disconnect websocket to SPC Web Gateway."""
        _LOGGER.debug(
            "Stop websocket communication with SPC Bridge: %s:%i",
            self._gw_ip_address,
            self._gw_port,
        )
        if self._websocket:
            self._websocket.stop()
        self._websocket = None

    def ws_state(self):
        if self._websocket:
            return self._websocket.state

    def _handle_output_event(self, event):
        output_id = int(event.get("mg_id", 0))
        event_id = int(event.get("ev_id", 0))
        state = None
        if event_id == 7013:
            state = True
        elif event_id == 7014:
            state = False

        if state is not None and output_id > 0:
            return self._set_value("output", output_id, {"state": state})
        return False

    def _handle_door_event(self, event):
        door_id = int(event.get("door_id", 0))
        event_id = int(event.get("ev_id", 0))
        username = event.get("user_name", "")

        if event_id == 3000:
            return self._set_value("door", door_id, {"entry_granted_user": username})

        if event_id == 3001:
            return self._set_value("door", door_id, {"entry_denied_user": username})
            return
        if event_id == 3002:
            return self._set_value("door", door_id, {"exit_granted_user": username})

        if event_id == 3003:
            return self._set_value("door", door_id, {"exit_denied_user": username})
            return

        mode = None
        if event_id == 3007:  # Normal mode
            mode = 0
        elif event_id == 3008:  # Locked mode
            mode = 1
        elif event_id == 3009:  # Unlocked mode
            mode = 2

        if mode is not None and door_id > 0:
            return self._set_value("door", door_id, {"mode": mode})
        return False

    async def _handle_system_event(self, event):
        # Reload and Update all SPC data after Engineer logout
        await self._bridge.async_reload_config()
        return

    def _handle_area_event(self, event):
        event_id = int(event.get("ev_id", 0))
        area_id = int(event.get("area_id", 0))
        sia_code = event.get("sia_code", "")
        username = event.get("user_name", "")

        if area_id > 0:
            if sia_code == "OG":  # Unset
                # Arm mode 0 and disarm username
                self._set_value(
                    "area",
                    area_id,
                    {
                        "mode": 0,
                        "unset_user": username,
                        "changed_by": username,
                        "pending_exit": False,
                    },
                )
                self._set_value("panel", None, {"changed_by": username})
                return
            elif sia_code == "NL":
                if event_id == 3502:  # Partset A
                    self._set_value(
                        "area",
                        area_id,
                        {"mode": 1, "changed_by": username, "pending_exit": False},
                    )
                    self._set_value("panel", None, {"changed_by": username})
                    return
                elif event_id == 3503:  # Partset B
                    self._set_value(
                        "area",
                        area_id,
                        {"mode": 2, "changed_by": username, "pending_exit": False},
                    )
                    self._set_value("panel", None, {"changed_by": username})
                    return
            elif sia_code == "CG":  # Fullset
                self._set_value(
                    "area",
                    area_id,
                    {
                        "mode": 3,
                        "set_user": username,
                        "changed_by": username,
                        "pending_exit": False,
                    },
                )
                self._set_value("panel", None, {"changed_by": username})
                return

        return

    async def _async_handle_zone_event(self, event):
        event_id = int(event.get("ev_id", 0))
        zone_id = int(event.get("zone_id", 0))

        if (
            (event_id >= 1000 and event_id <= 1029)
            or (event_id >= 1100 and event_id <= 1129)
            or (event_id >= 1200 and event_id <= 1223)
            or (event_id >= 1300 and event_id <= 1323)
            or (event_id >= 1400 and event_id <= 1423)
            or (event_id >= 1500 and event_id <= 1523)
            or (event_id >= 1600 and event_id <= 1623)
            or (event_id >= 1700 and event_id <= 1723)
            or (event_id >= 1800 and event_id <= 1823)
            or (event_id >= 1900 and event_id <= 1923)
            or (event_id >= 2000 and event_id <= 2023)
            or (event_id >= 2100 and event_id <= 2117)
            or (event_id >= 8004 and event_id <= 8005)
        ):
            if zone_id > 0:
                try:
                    data = await self._http_client.async_get_zones(id=zone_id)
                    if data and isinstance(data, list):
                        zone = data[0]
                        id = zone.get("id", 0)
                        input = zone.get("input", 0)
                        status = zone.get("status", 0)
                        if id > 0:
                            self._set_value(
                                "zone", id, {"input": input, "status": status}
                            )
                            return True

                    return False

                except Exception as err:
                    _LOGGER.warning("SPC EXCEPTION: %s", err)
                    return False

    async def _async_ws_handler(self, data):
        """Process incoming websocket message."""
        event = data["data"].get("sia") or data["data"].get("event")
        if event is None:
            return

        self._set_value("panel", None, {"event": json.dumps(event)})

        event_id = int(event.get("ev_id", 0))
        sia_code = event.get("sia_code", "")

        # Output events
        if event_id == 7013 or event_id == 7014:
            self._handle_output_event(event)
            return

        # Door events
        if (
            event_id == 3000
            or event_id == 3001
            or event_id == 3002
            or event_id == 3003
            or event_id == 3007
            or event_id == 3008
            or event_id == 3009
        ):
            self._handle_door_event(event)
            return

        # System events
        if sia_code == "LX":
            await self._handle_system_event(event)
            return

        # Area events
        if sia_code == "CI" or sia_code == "CG" or sia_code == "OG" or sia_code == "NL":
            self._handle_area_event(event)
            return

        # Zone events
        await self._async_handle_zone_event(event)
        return



"""Exceptions"""


class SpcException(Exception):
    """Base error for SPC."""


class RequestError(SpcException):
    """Unable to fulfill request.

    Raised when device cannot be reached.
    """


class ResponseError(SpcException):
    """Invalid response."""


class Unauthorized(SpcException):
    """Username is not authorized."""


class Forbidden(SpcException):
    """Endpoint is not accessible due to low permissions."""


class LoginRequired(SpcException):
    """User is logged out."""


class MethodNotAllowed(SpcException):
    """Invalid request."""


class PathNotFound(SpcException):
    """Path not found."""


ERRORS = {
    401: Unauthorized,
    403: Forbidden,
    404: PathNotFound,
    405: MethodNotAllowed,
}


def raise_error(error: int) -> None:
    """Raise error."""
    cls = ERRORS.get(error, SpcException)
    raise cls(error)



import logging

from .const import DoorMode
from .spc_error import SpcError
from .utils import _load_enum

_LOGGER = logging.getLogger(__name__)


class Door:
    """Represents a SPC door lock."""

    def __init__(self, bridge, spc_door):
        self._bridge = bridge
        self._id = spc_door.get("id")
        self._name = spc_door.get("name")
        self._values = {
            "mode": _load_enum(DoorMode, spc_door.get("mode", 0)),
            "entry_granted_user": "",
            "entry_denied_user": "",
            "exit_granted_user": "",
            "exit_denied_user": "",
        }
        self._http_client = bridge._http_client

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def entry_granted(self):
        return self._values["entry_granted_user"]

    @property
    def entry_denied(self):
        return self._values["entry_denied_user"]

    @property
    def exit_granted(self):
        return self._values["exit_granted_user"]

    @property
    def exit_denied(self):
        return self._values["exit_denied_user"]

    @property
    def mode(self):
        return self._values["mode"]

    def change_values(self, values) -> list:
        changed_values = []

        for k, v in values.items():
            value = v
            if k == "mode":
                value = _load_enum(DoorMode, v)
            if value != self._values[k]:
                self._values[k] = value
                changed_values.append(k)

        return changed_values

    async def async_command(self, command, code) -> dict:
        if code is None:
            return SpcError(54).error

        username = None
        password = None
        username, password = self._bridge.get_user_credentials(code)
        if username is None or password is None:
            return SpcError(54).error

        return await self._http_client.async_command_door(
            command, self._id, username, password
        )



"""Python wrapper for the Lundix SPC Bridge REST API."""

import asyncio
import logging

from .area import Area
from .const import ZoneInput
from .door import Door
from .output import Output
from .panel import Panel
from .spc_http_client import SpcHttpClient
from .spc_ws_client import SpcWsClient
from .user import User
from .zone import Zone

_LOGGER = logging.getLogger(__name__)

__all__ = ["Area", "Door", "Output", "Panel", "User", "Zone", "ZoneInput"]


class SpcBridge:
    """Alarm system representation."""

    def __init__(
        self,
        gw_ip_address,
        gw_port,
        credentials,
        users_config,
        loop,
        session,
        http_client,
        async_callback,
    ):
        """Initialize the client."""
        self._async_callback = async_callback
        self._panel = None
        self._users = {}
        self._areas = {}
        self._zones = {}
        self._outputs = {}
        self._doors = {}
        self._users_config = users_config

        self._http_client = SpcHttpClient(
            gw_ip_address=gw_ip_address,
            gw_port=gw_port,
            credentials=credentials,
            http_client=http_client,
        )

        self._ws_client = SpcWsClient(
            bridge=self,
            gw_ip_address=gw_ip_address,
            gw_port=gw_port,
            credentials=credentials,
            loop=loop,
            session=session,
            http_client=self._http_client,
        )

    @property
    def info(self):
        """Retrieve basic panel info."""
        return self._info

    @property
    def panel(self):
        """Retrieve panel data."""
        return self._panel

    @property
    def users(self):
        """Retrieve all available users."""
        return self._users

    @property
    def areas(self):
        """Retrieve all available areas."""
        return self._areas

    @property
    def zones(self):
        """Retrieve all available zones."""
        return self._zones

    @property
    def outputs(self):
        """Retrieve all available outputs (mapping gates)."""
        return self._outputs

    @property
    def doors(self):
        """Retrieve all available doors."""
        return self._doors

    def ws_start(self):
        """Start Websocket communication"""
        self._ws_client.ws_start()

    def ws_stop(self):
        """Stop Websocket communication"""
        self._ws_client.ws_stop()

    def set_value(self, resource, id, values):
        if resource == "panel":
            if self._panel.change_values(values):
                asyncio.create_task(
                    self._async_callback("update", self._panel.id, [self._panel])
                )

        if resource == "area":
            area = self._areas[id]
            if area:
                if area.change_values(values):
                    asyncio.create_task(
                        self._async_callback(
                            "update", self._panel.id, [area, self._panel]
                        )
                    )

        if resource == "zone":
            zone = self._zones[id]
            if zone:
                changed_values = zone.change_values(values)
                if "input" in changed_values and "status" not in changed_values:
                    # Only zone is needed to update
                    asyncio.create_task(
                        self._async_callback("update", self._panel.id, [zone])
                    )
                elif changed_values:
                    # Update zone, area of zone and panel
                    asyncio.create_task(
                        self._async_callback(
                            "update", self._panel.id, [zone, zone._area, self._panel]
                        )
                    )

        if resource == "output":
            output = self._outputs[id]
            if output:
                if output.change_values(values):
                    asyncio.create_task(
                        self._async_callback("update", self._panel.id, [output])
                    )

        if resource == "door":
            door = self._doors[id]
            if door:
                if door.change_values(values):
                    asyncio.create_task(
                        self._async_callback("update", self._panel.id, [door])
                    )

    def get_user_credentials(self, code):
        """Parse a user code and return user's name and password"""
        if self._users_config:
            for u in self._users.values():
                if code == u.ha_pincode:
                    return u.name, u.spc_password
        else:
            spc_len = self._panel.pincode_length
            code_len = len(code)
            if code_len > spc_len:
                userid = code[: (code_len - spc_len)]
                if u := self._users.get(int(userid)):
                    return u.name, code[-spc_len:]
        return None, None

    async def async_load_config(self):
        """Fetch configuration from SPC to initialize."""

        _LOGGER.debug("Load SPC configuration")
        try:
            panel_data = await self._http_client.async_get_panel()
            await asyncio.sleep(0.1)
            users_data = await self._http_client.async_get_users()
            await asyncio.sleep(0.1)
            areas_data = await self._http_client.async_get_areas()
            await asyncio.sleep(0.1)
            zones_data = await self._http_client.async_get_zones()
            await asyncio.sleep(0.1)
            outputs_data = await self._http_client.async_get_outputs()
            await asyncio.sleep(0.1)
            doors_data = await self._http_client.async_get_doors()

            # Get exit and entry times for each area
            for a in areas_data:
                a["exittime"] = 0
                a["entrytime"] = 0
                if (id := a.get("id")) is not None:
                    await asyncio.sleep(0.1)
                    config = await self._http_client.async_get_area_configs(id=id)
                    if config and list(config):
                        a["exittime"] = config[0].get("exittime", 0)
                        a["entrytime"] = config[0].get("entrytime", 0)

        except Exception as err:
            _LOGGER.warning("Exception trying to get SPC data: %s", err)
            raise

        if not zones_data or not areas_data:
            return False

        """ Create class objects """
        for a in areas_data:
            area = Area(self, a)
            area_zones = [
                Zone(self, area, z)
                for z in zones_data
                if z.get("area_id") == a.get("id")
            ]
            area.zones = area_zones
            self._areas[area.id] = area
            self._zones.update({z.id: z for z in area_zones})

        self._panel = Panel(self, panel_data, self._areas.values())

        for u in users_data:
            user = User(u, self._users_config)
            self._users[user.id] = user

        for o in outputs_data:
            output = Output(self, o)
            self._outputs[output.id] = output

        for d in doors_data:
            door = Door(self, d)
            self._doors[door.id] = door

        return True

    async def async_reload_config(self):
        """Reload configuration from SPC."""
        _LOGGER.debug("Reload SPC configuration")
        asyncio.create_task(self._async_callback("reload", self._panel.id))
        return True

    async def test_connection(self):
        """Test that communication parameters are valid"""
        try:
            panel_data = await self._http_client.async_get_panel()
            await asyncio.sleep(0.1)
            users_data = await self._http_client.async_get_users()
            await asyncio.sleep(0.1)
            areas_data = await self._http_client.async_get_areas()
            await asyncio.sleep(0.1)
            zones_data = await self._http_client.async_get_zones()
            await asyncio.sleep(0.1)
            outputs_data = await self._http_client.async_get_outputs()
            await asyncio.sleep(0.1)
            doors_data = await self._http_client.async_get_doors()
            await asyncio.sleep(0.1)

            spc_data = {
                "panel": panel_data,
                "users": users_data,
                "areas": areas_data,
                "zones": zones_data,
                "outputs": outputs_data,
                "doors": doors_data,
            }

            self.ws_start()
            for _ in range(10):
                await asyncio.sleep(1)
                if self._ws_client.ws_state() == "running":
                    self.ws_stop()
                    return spc_data

            self.ws_stop()
            return None
        except Exception:
            self.ws_stop()
            return None

    async def async_command(self, command, id) -> int:
        try:
            if command == "clear_alerts":
                return await self._http_client.async_command_clear_alerts()
            else:
                return await self._http_client.async_command_area(command, id)
        except Exception:
            raise

    async def async_get_arm_status(self, arm_mode, id=None) -> dict:
        try:
            return await self._http_client.async_get_area_arm_status(arm_mode, id)
        except Exception:
            raise



import logging

from .spc_error import SpcError

_LOGGER = logging.getLogger(__name__)


class Output:
    """Represents a SPC output/mapping gate."""

    def __init__(self, bridge, spc_output):
        self._bridge = bridge
        self._id = spc_output.get("id")
        self._name = spc_output.get("name")
        self._values = {
            "state": spc_output.get("state", 0) == 1,
        }
        self._http_client = bridge._http_client

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def state(self) -> bool:
        return self._values["state"]

    def change_values(self, values) -> list:
        changed_values = []

        if values.get("state") is not None:
            if values["state"] != self._values["state"]:
                self._values["state"] = values["state"]
                changed_values.append("state")

        return changed_values

    async def async_command(self, command, code) -> dict:
        if code is None:
            return SpcError(54).error

        username = None
        password = None
        username, password = self._bridge.get_user_credentials(code)
        if username is None or password is None:
            return SpcError(54).error

        return await self._http_client.async_command_output(
            command, self._id, username, password
        )



import json
import logging
from asyncio import ensure_future

import aiohttp

_LOGGER = logging.getLogger(__name__)

STATE_STARTING = "starting"
STATE_RUNNING = "running"
STATE_STOPPED = "stopped"

RETRY_TIMER = 15


class AIOWSClient:
    """Websocket transport, session handling, message generation."""

    """https://github.com/Kane610/deconz/blob/master/pydeconz/websocket.py"""

    def __init__(self, loop, session, url, async_callback):
        """Create resources for websocket communication."""
        self._loop = loop
        self._session = session
        self._url = url
        self._async_callback = async_callback
        self._data = None
        self._state = None
        self._task = None
        self._retry_timer = None

    @property
    def data(self):
        return self._data

    @property
    def state(self):
        """"""
        return self._state

    @state.setter
    def state(self, value):
        """"""
        self._state = value
        _LOGGER.debug("Websocket %s", value)

    def start(self):
        if self.state != STATE_RUNNING:
            self.state = STATE_STARTING
        if self._retry_timer:
            self._retry_timer.cancel()
            self._retry_timer = None
        self._task = self._loop.create_task(self.running())

    async def running(self):
        """Start websocket connection."""
        try:
            async with self._session.ws_connect(self._url, receive_timeout=30) as ws:
                self.state = STATE_RUNNING
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        ensure_future(self._async_callback(json.loads(msg.data)))
                        _LOGGER.debug("Websocket data: %s", msg.data)
                    elif (
                        msg.type == aiohttp.WSMsgType.CLOSED
                        or msg.type == aiohttp.WSMsgType.ERROR
                    ):
                        break
        except aiohttp.ClientConnectorError:
            if self.state != STATE_STOPPED:
                self.retry()
        except Exception as err:
            _LOGGER.error("Websocket timeout/error %s", err)
            if self.state != STATE_STOPPED:
                self.retry()
        else:
            if self.state != STATE_STOPPED:
                self.retry()

    def stop(self):
        """Close websocket connection."""
        self.state = STATE_STOPPED
        if self._task:
            self._task.cancel()
        if self._retry_timer:
            self._retry_timer.cancel()
            self._retry_timer = None

    def retry(self):
        """Retry to connect to SPC."""
        self.state = STATE_STARTING
        self._retry_timer = self._loop.call_later(RETRY_TIMER, self.start)
        _LOGGER.warning("Reconnecting to SPC in %i.", RETRY_TIMER)



import logging

_LOGGER = logging.getLogger(__name__)


class User:
    """Represents a SPC user."""

    def __init__(self, spc_user, users_config):
        self._id = spc_user.get("id")
        self._name = spc_user.get("name")
        if users_config and users_config.get(str(self._id)):
            self._ha_pincode = users_config[str(self._id)].get("ha_pincode", "")
            self._spc_password = users_config[str(self._id)].get("spc_password", "")
        else:
            self._ha_pincode = ""
            self._spc_password = ""

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def ha_pincode(self):
        return self._ha_pincode

    @property
    def spc_password(self):
        return self._spc_password



from enum import Enum


class ArmMode(Enum):
    UNSET = 0
    PART_SET_A = 1
    PART_SET_B = 2
    FULL_SET = 3
    PARTLY_SET_A = 97
    PARTLY_SET_B = 98
    PARTLY_FULL_SET = 99


class ZoneInput(Enum):
    CLOSED = 0
    OPEN = 1
    SHORT = 2
    DISCONNECTED = 3
    PIR_MASKED = 4
    DC_SUBSTITUTION = 5
    SENSOR_MISSING = 6
    OFFLINE = 7


class ZoneType(Enum):
    ALARM = 0
    ENTRY_EXIT = 1
    EXIT_TERMINATOR = 2
    FIRE = 3
    FIRE_EXIT = 4
    LINE = 5
    PANIC = 6
    HOLD_UP = 7
    TAMPER = 8
    TECHNICAL = 9
    MEDICAL = 10
    KEYARM = 11
    UNUSED = 12
    SHUNT = 13
    X_SHUNT = 14
    FAULT = 15
    LOCK_SUPERVISION = 16
    SEISMIC = 17
    ALL_OKAY = 18
    HOLDUP_FAULT = 19
    WARNING_FAULT = 20
    SETTING_AUTHORISATION = 21
    LOCK_ELEMENT = 22
    GLASSBREAK = 23
    WATER = 24
    HEAT = 25
    FRIDGE_FREEZER = 26
    GAS = 27
    SPRINKLER = 28
    CO = 29
    ENTRY_EXIT_2 = 30


class ZoneStatus(Enum):
    OK = 0
    INHIBIT = 1
    ISOLATE = 2
    SOAK = 3
    TAMPER = 4
    ALARM = 5
    OK_NOT_RECENT = 6
    TROUBLE = 7


class DoorMode(Enum):
    NORMAL = 0
    LOCKED = 1
    UNLOCKED = 2



import logging

from .const import ArmMode
from .spc_error import SpcError

_LOGGER = logging.getLogger(__name__)


class Panel:
    """Represents a SPC panel."""

    def __init__(self, bridge, panel_data, areas):
        self._bridge = bridge
        self._id = panel_data.get("serial")
        self._type = panel_data.get("type")
        self._model = panel_data.get("model")
        self._serial = panel_data.get("serial")
        self._firmware = panel_data.get("firmware")
        self._pincode_length = int(panel_data.get("pincode_length", 0))
        self._http_client = bridge._http_client
        self._values = {
            "event": "",
            "changed_by": "",
        }
        self._areas = areas

    @property
    def id(self):
        return self._id

    @property
    def type(self):
        return self._type

    @property
    def model(self):
        return self._model

    @property
    def serial(self):
        return self._serial

    @property
    def firmware(self):
        return self._firmware

    @property
    def pincode_length(self):
        return self._pincode_length

    @property
    def a_enabled(self):
        for a in self._areas:
            if a.a_enabled:
                return a.a_enabled
        return False

    @property
    def a_name(self):
        for a in self._areas:
            if a.a_name is not None:
                return a.a_name
        return ""

    @property
    def b_enabled(self):
        for a in self._areas:
            if a.b_enabled:
                return a.b_enabled
        return False

    @property
    def b_name(self):
        for a in self._areas:
            if a.b_name is not None:
                return a.b_name
        return ""

    @property
    def exittime(self):
        _exittime = 0
        for a in self._areas:
            if a.exittime > _exittime:
                _exittime = a.exittime
        return _exittime

    @property
    def entrytime(self):
        _entrytime = 0
        for a in self._areas:
            if a.entrytime > _entrytime:
                _entrytime = a.entrytime
        return _entrytime

    @property
    def event(self) -> str:
        return self._values.get("event", "")

    @property
    def changed_by(self) -> str:
        return self._values.get("changed_by", "")

    @property
    def values(self):
        return self._values

    def change_values(self, values) -> list:
        changed_values = []
        if values.get("event") is not None:
            self._values["event"] = values["event"]
            changed_values.append("event")
        if values.get("changed_by") is not None:
            self._values["changed_by"] = values["changed_by"]
            changed_values.append("changed_by")
        return changed_values

    @property
    def mode(self):
        _mode = None
        for a in self._areas:
            _area_mode = a.mode
            if _mode is None:
                _mode = _area_mode
            elif _area_mode.value != _mode.value:
                _m = _mode
                if _area_mode.value > _mode.value:
                    _m = _area_mode
                if _m == ArmMode.PART_SET_A:
                    _mode = ArmMode.PARTLY_SET_A
                if _m == ArmMode.PART_SET_B:
                    _mode = ArmMode.PARTLY_SET_B
                if _m == ArmMode.FULL_SET:
                    _mode = ArmMode.PARTLY_FULL_SET

        return _mode

    @property
    def pending_exit(self):
        _state = False
        for a in self._areas:
            if a.pending_exit is True:
                _state = a.pending_exit
        return _state

    @property
    def alarm_status(self):
        _alarm_status = {
            "intrusion": False,
            "fire": False,
            "tamper": False,
            "problem": False,
            "verified": False,
        }

        for a in self._areas:
            _area_alarm_status = a.alarm_status
            for k in _area_alarm_status:
                if _area_alarm_status[k] is True:
                    _alarm_status[k] = True

        return _alarm_status

    @property
    def intrusion(self) -> bool:
        return self.alarm_status.get("intrusion", False)

    @property
    def fire(self) -> bool:
        return self.alarm_status.get("fire", False)

    @property
    def tamper(self) -> bool:
        return self.alarm_status.get("tamper", False)

    @property
    def problem(self) -> bool:
        return self.alarm_status.get("problem", False)

    @property
    def verified(self) -> bool:
        return self.alarm_status.get("verified", False)

    async def async_command(self, command, code) -> int:
        username = None
        password = None
        if code is not None:
            username, password = self._bridge.get_user_credentials(code)
            if username is None or password is None:
                return SpcError(54).error

        if command == "clear_alerts":
            return await self._http_client.async_command_clear_alerts(
                username, password
            )
        else:
            err = await self._http_client.async_command_area(
                command, None, username, password
            )
            if command == "set_delayed":
                if isinstance(err, dict) and err.get("code", 0) > 0:
                    return err
                if isinstance(err, list):
                    for e in err.values():
                        if e.get("code", 0) > 0:
                            return err
                for a in self._areas:
                    self._bridge.set_value("area", a.id, {"pending_exit": True})
            else:
                for a in self._areas:
                    self._bridge.set_value("area", a.id, {"pending_exit": False})
            return err

    async def async_get_arm_status(self, arm_mode) -> dict:
        return await self._http_client.async_get_area_arm_status(arm_mode)



"""SPC http client"""

import logging

import httpx

from .exceptions import RequestError, raise_error
from .spc_error import SpcError

_LOGGER = logging.getLogger(__name__)


class SpcHttpClient:
    def __init__(self, gw_ip_address, gw_port, credentials, http_client):
        self._gw_ip_address = gw_ip_address
        self._gw_port = gw_port
        self._credentials = credentials
        self._http_client = http_client

    async def _async_request(self, http_client, _auth, url):
        try:
            if _auth:
                response = await http_client(
                    url,
                    auth=_auth,
                    timeout=5,
                )
            else:
                response = await http_client(
                    url,
                    timeout=5,
                )

            if response.status_code != 200 and response.status_code != 404:
                _LOGGER.error(
                    "Error getting/settings SPC data from/to %s, HTTP status %d",
                    url,
                    response.status_code,
                )
                raise_error(response.status_code)

            result = response.json()

        except httpx.TimeoutException as errt:
            message = f"Timeout getting/setting SPC data from/to {url}."
            _LOGGER.error(message)
            raise RequestError(message) from errt

        except (httpx.TransportError, httpx.HTTPStatusError) as errc:
            message = f"Error getting/setting SPC data from/to {url}."
            _LOGGER.error(message)
            raise RequestError(message) from errc

        except httpx.RequestError as err:
            message = f"Error getting/setting SPC data from/to {url}."
            _LOGGER.error(message)
            raise RequestError(message) from err

        return result

    async def _async_get_data(self, resource, id=None):
        """Get the SPC data for the resource. (HTTP GET)"""
        proto = "http"
        auth = None
        if len(self._credentials["get_username"]) and len(
            self._credentials["get_password"]
        ):
            proto = "https"
            auth = httpx.DigestAuth(
                username=self._credentials["get_username"],
                password=self._credentials["get_password"],
            )

        if id is not None:
            url = f"{proto}://{self._gw_ip_address}:{self._gw_port}/spc/{resource}/{id}"
        else:
            url = f"{proto}://{self._gw_ip_address}:{self._gw_port}/spc/{resource}"

        try:
            data = await self._async_request(self._http_client.get, auth, url)
            if not data:
                _LOGGER.error("Unable to get data from SPC")
                return None
        except Exception:
            _LOGGER.error("Unable to get data from SPC")
            raise

        return data

    async def _async_put_data(
        self, resource, command, id=None, username=None, password=None
    ):
        """Send a command to SPC (HTTP PUT)."""
        proto = "http"
        auth = None
        if len(self._credentials["put_username"]) and len(
            self._credentials["put_password"]
        ):
            proto = "https"
            auth = httpx.DigestAuth(
                username=self._credentials["put_username"],
                password=self._credentials["put_password"],
            )

        if id is not None:
            url = f"{proto}://{self._gw_ip_address}:{self._gw_port}/spc/{resource}/{id}/{command}"
        else:
            url = f"{proto}://{self._gw_ip_address}:{self._gw_port}/spc/{resource}/{command}"

        if username and password:
            url = f"{url}?username={username}&password={password}"

        try:
            data = await self._async_request(self._http_client.put, auth, url)
            if not data:
                _LOGGER.error("Unable to send a command to SPC")
                return None
        except Exception:
            _LOGGER.error("Unable to to send a command to SPC")
            raise

        return data

    def _validate_header(self, reply) -> int:
        if (
            not reply
            or (
                reply.get("status", "") != "success"
                and reply.get("status", "") != "error"
            )
            or not reply.get("data")
        ):
            return 998
        if (err := int(reply["data"].get("code", -1))) > 0:
            return err
        if reply.get("status", "") == "error":
            return 998
        return 0

    async def async_get_users(self, id=None):
        """Get Users data"""

        def _normalize(d):
            return {
                "id": int(d.get("user_id", 0)),
                "name": d.get("name", ""),
            }

        data = await self._async_get_data("user", id)
        if data and data.get("status", "") == "success" and data.get("data"):
            users = data["data"].get("user_config")
            if isinstance(users, dict):
                return [_normalize(u) for u in [users]]
            elif isinstance(users, list):
                return [_normalize(u) for u in users]
        return {}

    async def async_get_panel(self):
        """Get Panel data"""
        data = await self._async_get_data("panel")
        if data and data.get("status", "") == "success" and data.get("data"):
            panel = data["data"].get("panel_summary")
            if panel:
                return {
                    "type": panel.get("spc_type", ""),
                    "model": panel.get("spc_variant", ""),
                    "serial": panel.get("spc_serial_no", ""),
                    "firmware": panel.get("spc_fw_version", ""),
                    "pincode_length": panel.get("pin_digits", 4),
                }
            else:
                return None

    async def async_get_area_configs(self, id=None):
        """Get Area config"""

        def _normalize(a):
            return {
                "id": int(a.get("id", 0)),
                "entrytime": int(a.get("entrytime", 0)),
                "exittime": int(a.get("exittime", 0)),
            }

        if id is not None:
            data = await self._async_get_data(f"area/{id}/config")
        else:
            data = await self._async_get_data("area/config")

        if data and data.get("status", "") == "success" and data.get("data"):
            areas = data["data"].get("area")
            if isinstance(areas, dict):
                return [_normalize(a) for a in [areas]]
            elif isinstance(areas, list):
                return [_normalize(a) for a in areas]

        return {}

    async def async_get_area_arm_status(self, arm_mode, id=None):
        """Get Area arm status"""

        def _normalize(a):
            reasons = []
            for x in range(100):
                if (reason := a.get(f"reason_{x}")) is not None:
                    r = int(reason)
                    if r > 100 and r < 999:
                        reasons.append(f"area_{r-100}")
                    elif r > 1000 and r < 1999:
                        reasons.append(f"zone_{r-1000}")
                    elif r == 10006:
                        reasons.append("engineer_on_site")
                    elif r == 10007:
                        reasons.append("no_rights")
                    elif r != 0:
                        reasons.append("undefined")
                else:
                    break
            return {
                "area_id": int(a.get("area_id", 0)),
                "reasons": reasons,
            }

        if id is not None:
            data = await self._async_get_data(f"area/{id}/{arm_mode}_status")
        else:
            data = await self._async_get_data(f"area/{arm_mode}_status")

        if (
            data
            and data.get("status", "") == "success"
            and data.get("data")
            and (d := data.get("data").get("reply_get_area_change_mode_status"))
        ):
            areas = d.get("area_change_mode_status")
            if isinstance(areas, dict):
                return [_normalize(a) for a in [areas]]
            elif isinstance(areas, list):
                return [_normalize(a) for a in areas]
        return {}

    async def async_get_areas(self, id=None):
        """Get Area data"""

        def _normalize(a):
            return {
                "id": int(a.get("area_id", 0)),
                "name": a.get("area_name", ""),
                "mode": int(a.get("mode", 0)),
                "a_enabled": int(a.get("partseta_enable", 0)) == 1,
                "a_name": a.get("partseta_name", "") or "",
                "b_enabled": int(a.get("partsetb_enable", 0)) == 1,
                "b_name": a.get("partsetb_name", ""),
                "set_user": a.get("last_set_user_name", ""),
                "unset_user": a.get("last_unset_user_name", ""),
            }

        data = await self._async_get_data("area", id)
        if data and data.get("status", "") == "success" and data.get("data"):
            areas = data["data"].get("area_status")
            if isinstance(areas, dict):
                return [_normalize(a) for a in [areas]]
            elif isinstance(areas, list):
                return [_normalize(a) for a in areas]
        return {}

    async def async_get_zones(self, id=None):
        """Get Zone data"""

        def _normalize(z):
            return {
                "id": int(z.get("zone_id", 0)),
                "name": z.get("zone_name", ""),
                "type": int(z.get("type", 0)),
                "input": int(z.get("logic_input", 0)),
                "status": int(z.get("status", 0)),
                "area_id": int(z.get("area_id", 0)),
                "area_name": z.get("area_name", ""),
            }

        data = await self._async_get_data("zone", id)
        if data and data.get("status", "") == "success" and data.get("data"):
            zones = data["data"].get("zone_status")
            if isinstance(zones, dict):
                return [_normalize(z) for z in [zones]]
            elif isinstance(zones, list):
                return [_normalize(z) for z in zones]
        return {}

    async def async_get_outputs(self, id=None):
        """Get Output data"""

        def _normalize(o):
            return {
                "id": int(o.get("mg_id", 0)),
                "name": o.get("mg_name", ""),
                "state": int(o.get("state", 0)),
            }

        data = await self._async_get_data("output", id)
        if data and data.get("status", "") == "success" and data.get("data"):
            outputs = data["data"].get("mg_status")
            if isinstance(outputs, dict):
                return [_normalize(o) for o in [outputs]]
            elif isinstance(outputs, list):
                return [_normalize(o) for o in outputs]
        return {}

    async def async_get_doors(self, id=None):
        """Get Door data"""

        def _normalize(d):
            return {
                "id": int(d.get("door_id", 0)),
                "name": d.get("zone_name", ""),
                "status": int(d.get("status", 0)),
                "mode": int(d.get("mode", 0)),
            }

        data = await self._async_get_data("door", id)
        if data and data.get("status", "") == "success" and data.get("data"):
            doors = data["data"].get("door_status")
            if isinstance(doors, dict):
                return [_normalize(d) for d in [doors]]
            elif isinstance(doors, list):
                return [_normalize(d) for d in doors]
        return {}

    async def async_command_area(
        self, command, id=None, username=None, password=None
    ) -> dict:
        """Area commands"""
        try:
            reply = await self._async_put_data("area", command, id, username, password)
            if (err := self._validate_header(reply)) > 0:
                return SpcError(err).error

            _data = reply.get("data")
            if (
                _data.get("reply_area_change_mode")
                and (err := int(_data.get("reply_area_change_mode").get("result", -1)))
                > 0
            ):
                return SpcError(err).error
            if _area_change_mode := _data.get("reply_area_change_mode").get(
                "area_change_mode"
            ):
                if isinstance(_area_change_mode, dict):
                    if (err := int(_area_change_mode.get("result", -1))) >= 0:
                        return SpcError(err).error
                elif isinstance(_area_change_mode, list):
                    errors = {}
                    for a in _area_change_mode:
                        if (id := int(a.get("area_id", -1)) > 0) and (
                            err := int(a.get("result", -1)) >= 0
                        ):
                            errors[id] = SpcError(err).error
                        else:
                            return SpcError(998).error
                    return errors

            return SpcError(998).error
        except Exception as err:
            return SpcError(998).error

    async def async_command_zone(
        self, command, id=None, username=None, password=None
    ) -> dict:
        """Zone commands"""
        try:
            reply = await self._async_put_data("zone", command, id, username, password)
            if (err := self._validate_header(reply)) > 0:
                return SpcError(err).error

            _data = reply.get("data")
            if (
                _data.get("reply_zone_control")
                and (err := int(_data.get("reply_zone_control").get("result", -1))) > 0
            ):
                return SpcError(err).error
            if (
                _data.get("reply_zone_control").get("zone_control")
                and (
                    err := int(
                        _data.get("reply_zone_control")
                        .get("zone_control")
                        .get("result", -1)
                    )
                )
                >= 0
            ):
                return SpcError(err).error
            else:
                return SpcError(998).error
        except Exception as err:
            return SpcError(998).error

    async def async_command_output(
        self, command, id=None, username=None, password=None
    ):
        """Output commands"""
        try:
            reply = await self._async_put_data(
                "output", command, id, username, password
            )
            if (err := self._validate_header(reply)) > 0:
                return SpcError(err).error

            _data = reply.get("data")
            if (
                _data.get("reply_mg_control")
                and (err := int(_data.get("reply_mg_control").get("result", -1))) > 0
            ):
                return SpcError(err).error
            if (
                _data.get("reply_mg_control").get("mg_control")
                and (
                    err := int(
                        _data.get("reply_mg_control")
                        .get("mg_control")
                        .get("result", -1)
                    )
                )
                >= 0
            ):
                return SpcError(err).error
            else:
                return SpcError(998).error
        except Exception as err:
            return SpcError(998).error

    async def async_command_door(self, command, id=None, username=None, password=None):
        """Door commands"""
        try:
            reply = await self._async_put_data("door", command, id, username, password)
            if (err := self._validate_header(reply)) > 0:
                return SpcError(err).error

            _data = reply.get("data")
            if (
                _data.get("reply_door_control")
                and (err := int(_data.get("reply_door_control").get("result", -1))) > 0
            ):
                return SpcError(err).error
            if (
                _data.get("reply_door_control").get("door_control_result")
                and (
                    err := int(
                        _data.get("reply_door_control")
                        .get("door_control_result")
                        .get("result", -1)
                    )
                )
                >= 0
            ):
                return SpcError(err).error
            else:
                return SpcError(998).error
        except Exception as err:
            return SpcError(998).error

    async def async_command_clear_alerts(self, username=None, password=None):
        """Command - Clear All Alerts"""
        try:
            reply = await self._async_put_data(
                "alert", "clear", username=username, password=password
            )
            if (err := self._validate_header(reply)) > 0:
                return SpcError(err).error

            _data = reply.get("data")
            if (
                _data.get("reply_clear_all_alerts")
                and (err := int(_data.get("reply_clear_all_alerts").get("result", -1)))
                >= 0
            ):
                return SpcError(err).error
            else:
                return SpcError(998).error
        except Exception as err:
            return SpcError(998).error
